import importlib
import inspect

from .utils import *

__all__ = ['RegistrationSystem', 'register', 'register_inline', 'generate_config', 'get_module']


class RegistrationSystem:
    def __init__(self):
        self._registry = {}
        self._paths = {}

    def _register(self, resource, module_name, module_type):
        assert (type(module_name) is str or module_name is None and
                type(module_type) is str or module_type is None)

        stack = inspect.stack()
        source = inspect.getframeinfo(stack[2][0]).filename
        source = os.path.realpath(source)

        if module_type is None:
            module_type = os.path.basename(os.path.dirname(source))

        if module_name is None:
            try:
                module_name = snake_case(resource.__name__)
            except AttributeError:
                raise TypeError('Could not infer the resource name. Please specify it explicitly.'
                                '\nCheck the source file {}'.format(source)) from None

        modules = self._registry.setdefault(module_type, {})
        if module_name in modules and modules[module_name] != resource:
            raise ValueError('Trying to register another resource with the same name and type: '
                             '{}:{}'.format(module_type, module_name))
        modules[module_name] = resource
        self._paths.setdefault(source, set()).add((module_type, module_name))

        return resource

    def _analyze_file(self, path, source, config):
        config = [x for x in config if x['source'] != source]
        path = os.path.realpath(path)
        importlib.import_module(source)
        modules = self._paths.get(path, [])

        for module_type, module_name in modules:
            config.append({
                'module_type': module_type,
                'module_name': module_name,
                'source_path': path,
                'source': source,
            })

        return config

    def register(self, module_name: str = None, module_type: str = None):
        """
        A class/function decorator that registers a resource.

        Parameters
        ----------
        module_name: str, optional.
            The name of the module. If None, the __name__ attribute converted to snake_case is used.
        module_type: str, optional
            The type of the module. If None, the folder name it lies in is used.
        """

        def decorator(resource):
            return self._register(resource, module_name, module_type)

        return decorator

    def register_inline(self, resource, module_name: str = None, module_type: str = None):
        """
        Registers a resource.

        Parameters
        ----------
        resource
            Any object that needs to be registered.
        module_name: str, optional.
            The name of the module. If None, the __name__ attribute converted to snake_case is used.
        module_type: str, optional
            The type of the module. If None, the folder name it lies in is used.
        """

        return self._register(resource, module_name, module_type)

    def generate_config(self, root, db_path, upper_module, exclude=None):
        if exclude is None:
            exclude = []
        old_config, old_hashes = read_config(db_path)
        exclude = [os.path.abspath(os.path.join(root, x)) for x in exclude]
        sources_list = walk(root, upper_module, exclude)

        # cleaning up the deleted files
        current_paths = [os.path.realpath(x[0]) for x in sources_list]
        config = []
        for entry in old_config:
            source_path = entry['source_path']
            if source_path in current_paths:
                config.append(entry)
            else:
                # the file is deleted
                old_hashes.pop(source_path, None)

        hashes = {}
        for path, source in sources_list:
            hashes[path] = new_hash = get_hash(path)
            old_hash = old_hashes.get(path, '')
            if old_hash != new_hash:
                config = self._analyze_file(path, source, config)

        with open(db_path, 'w') as file:
            json.dump({'config': config, 'hashes': hashes}, file, indent=2)

    def get_module(self, module_type, module_name, db_path):
        config = read_config(db_path)[0]

        for entry in config:
            try:
                if entry['module_type'] == module_type and entry['module_name'] == module_name:
                    importlib.import_module(entry['source'])
                    return self._registry[entry['module_type']][entry['module_name']]
            except KeyError:
                handle_corruption(db_path)

        raise KeyError('The module "{}:{}" was not found'.format(module_type, module_name))


registration_system = RegistrationSystem()
register = registration_system.register
register_inline = registration_system.register_inline
generate_config = registration_system.generate_config
get_module = registration_system.get_module



import sys
import functools
import importlib
from typing import Callable

from .parser import parse_file, parse_string
from .helpers import Scope, LambdaFunction
from .tree_analysis import SyntaxTree
from .structures import *


class ResourceManager:
    """
    A config interpreter.

    Parameters
    ----------
    shortcuts: dict, optional
        a dict that maps keywords to paths. It is used to resolve paths during import.
    get_module: callable(module_type, module_name) -> object, optional
        a callable that loads external modules
    """

    def __init__(self, shortcuts: dict = None, get_module: Callable = None):
        self.get_module = get_module
        self._shortcuts = shortcuts or {}

        self._imported_configs = {}
        self._scopes = [Scope()]
        self._request_stack = []
        self._definitions_stack = []

    @classmethod
    def read_config(cls, source_path: str, shortcuts: dict = None, get_module: Callable = None):
        """
        Import the config located at `source_path` and return a ResourceManager instance.

        Parameters
        ----------
        source_path: str
            path to the config to import
        shortcuts: dict, optional
            a dict that maps keywords to paths. It is used to resolve paths during import.
        get_module: callable(module_type, module_name) -> object, optional
            a callable that loads external modules

        Returns
        -------
        resource_manager: ResourceManager
        """
        rm = cls(shortcuts, get_module)
        rm.import_config(source_path)
        return rm

    def import_config(self, path: str):
        """Import the config located at `path`."""
        path = self._resolve_path(path)
        result = self._import(path)
        self._update_resources(result)

    def string_input(self, source: str):
        """Interpret the `source`."""
        definitions, parents, imports = parse_string(source)
        result = self._get_resources(definitions, imports, parents, '')
        self._update_resources(result)

    def render_config(self) -> str:
        """
        Generate a string containing definitions of all the resources in the current scope.

        Returns
        -------
        config: str
        """
        assert len(self._scopes) == 1
        scope = self._scopes[0]
        result = ''
        for name, value in scope._undefined_resources.items():
            if type(value) is LazyImport:
                result += value.to_str(0)
        if result:
            result += '\n'

        for name, value in scope._undefined_resources.items():
            if type(value) is not LazyImport:
                result += '{} = {}\n\n'.format(name, value.to_str(0))

        return result[:-1]

    def save_config(self, path: str):
        """Render the config and save it to `path`."""
        with open(path, 'w') as file:
            file.write(self.render_config())

    def __getattr__(self, item):
        return self.get_resource(item)

    def get_resource(self, name: str):
        self._request_stack = []
        self._definitions_stack = []
        assert len(self._scopes) == 1
        return self._scopes[0].get_resource(name, self._render)

    def _update_resources(self, scope):
        assert len(self._scopes) == 1
        # TODO: what if some of the resources were already rendered?
        self._scopes[0].overwrite(scope)
        SyntaxTree.analyze(scope)

    def _import(self, absolute_path: str) -> Scope:
        if absolute_path in self._imported_configs:
            return self._imported_configs[absolute_path]
        # avoiding cycles
        self._imported_configs[absolute_path] = Scope()

        definitions, parents, imports = parse_file(absolute_path)
        result = self._get_resources(definitions, imports, parents, absolute_path)
        self._imported_configs[absolute_path] = result
        return result

    def _get_resources(self, definitions: List[Definition], imports: List[ImportPython], parents, absolute_path):
        # TODO: absolute_path is redundant
        parent_scope = Scope()
        for parent in parents:
            for path in parent.get_paths():
                path = self._resolve_path(path, absolute_path)
                parent_scope.overwrite(self._import(path))

        scope = Scope()
        for import_ in imports:
            for what, as_ in import_.values:
                if as_ is not None:
                    name = as_.body
                else:
                    name = what
                    packages = name.split('.')
                    if len(packages) > 1:
                        name = packages[0]
                scope.set_resource(name, LazyImport(import_.root, what, as_, import_.relative, import_.main_token))

        for definition in definitions:
            scope.set_resource(definition.name.body, definition.value)

        parent_scope.overwrite(scope)
        return parent_scope

    # TODO: change signature to path, source, shortcut
    def _resolve_path(self, path: str, source: str = ''):
        if path.count(':') > 1:
            raise SyntaxError('The path cannot contain more than one ":" separator.')

        parts = path.split(':', 1)
        if len(parts) > 1:
            shortcut, root = parts
            if shortcut not in self._shortcuts:
                message = 'Shortcut "%s" is not recognized' % shortcut
                if source:
                    message = 'Error while processing %s:\n ' % source + message
                raise ValueError(message)
            path = os.path.join(self._shortcuts[shortcut], root)
        else:
            path = os.path.join(os.path.dirname(source), path)

        path = os.path.expanduser(path)
        return os.path.realpath(path)

    def _render(self, node: Structure):
        self._definitions_stack.append(node)
        try:
            value = node.render(self)
        except BaseException as e:
            if not self._definitions_stack:
                raise
            # TODO: should all the traceback be printed?
            definition = self._definitions_stack[0]
            self._definitions_stack = []

            raise RuntimeError('An exception occurred while ' + definition.error_message() +
                               '\n    at %d:%d in %s' % definition.position()) from e
        self._definitions_stack.pop()
        return value

    def _render_lambda(self, node: Lambda):
        assert self._scopes
        return LambdaFunction(node, self._scopes[-1], self)

    def _render_resource(self, node: Resource):
        assert self._scopes
        return self._scopes[-1].get_resource(node.name.body, self._render)

    def _render_get_attribute(self, node: GetAttribute):
        data = self._render(node.target)
        return getattr(data, node.name.body)

    def _render_get_item(self, node: GetItem):
        target = self._render(node.target)
        args = tuple(self._render(arg) for arg in node.args)
        if not node.trailing_coma and len(args) == 1:
            args = args[0]
        return target[args]

    def _render_call(self, node: Call):
        target = self._render(node.target)
        args = []
        for vararg, arg in zip(node.varargs, node.args):
            temp = self._render(arg)
            if vararg:
                args.extend(temp)
            else:
                args.append(temp)
        kwargs = {param.name.body: self._render(param.value) for param in node.params}
        if node.lazy:
            return functools.partial(target, *args, **kwargs)
        return target(*args, **kwargs)

    def _render_module(self, node: Module):
        if self.get_module is None:
            raise ValueError('The function "get_module" was not provided, so your modules are unreachable')
        return self.get_module(node.module_type.body, node.module_name.body)

    def _render_literal(self, node: Literal):
        return eval(node.value.body)

    def _render_array(self, node: Array):
        return [self._render(x) for x in node.values]

    def _render_dictionary(self, node: Dictionary):
        return {self._render(key): self._render(value) for key, value in node.pairs}

    def _render_lazy_import(self, node: LazyImport):
        assert not node.relative
        if not node.from_:
            result = importlib.import_module(node.what)
            packages = node.what.split('.')
            if len(packages) > 1 and not node.as_:
                # import a.b.c
                return sys.modules[packages[0]]
            return result
        try:
            return getattr(importlib.import_module(node.from_), node.what)
        except AttributeError:
            pass
        try:
            return importlib.import_module(node.what, node.from_)
        except ModuleNotFoundError:
            return importlib.import_module(node.from_ + '.' + node.what)


read_config = ResourceManager.read_config



import os

from .tokenizer import TokenType, tokenize
from .structures import *


class Parser:
    def __init__(self, tokens):
        self.tokens = tokens
        self.position = 0

    def module(self):
        module_type = self.require(TokenType.IDENTIFIER)
        self.require(TokenType.COLON)
        name = self.require(TokenType.IDENTIFIER)

        return Module(module_type, name)

    def data(self):
        if self.matches(TokenType.IDENTIFIER):
            # module
            if self.matches(TokenType.COLON, shift=1):
                return self.module()
            else:
                # identifier
                return Resource(self.advance())

        if self.matches(TokenType.BRACKET_OPEN):
            return self.array()
        if self.matches(TokenType.DICT_OPEN):
            return self.dictionary()
        if self.matches(TokenType.LAMBDA):
            return self.lambda_()
        return Literal(self.require(TokenType.STRING, TokenType.NUMBER, TokenType.LITERAL))

    def lambda_(self):
        token = self.require(TokenType.LAMBDA)
        params = []
        if self.matches(TokenType.IDENTIFIER):
            params = [self.advance()]
            while self.ignore(TokenType.COMA):
                params.append(self.require(TokenType.IDENTIFIER))
        self.require(TokenType.COLON)
        return Lambda(params, self.expression(), token)

    def is_keyword(self):
        return self.matches(TokenType.IDENTIFIER) and self.matches(TokenType.EQUALS, shift=1)

    def params(self):
        self.require(TokenType.PAR_OPEN)
        lazy = self.ignore(TokenType.LAZY)
        vararg, args, keyword = [], [], []
        if self.ignore(TokenType.PAR_CLOSE):
            return args, vararg, keyword, lazy

        # if has positional
        if not self.is_keyword():
            vararg.append(self.ignore(TokenType.ASTERISK))
            args.append(self.expression())

            while self.ignore(TokenType.COMA):
                if self.is_keyword() or self.matches(TokenType.PAR_CLOSE):
                    break
                vararg.append(self.ignore(TokenType.ASTERISK))
                args.append(self.expression())

        # keyword
        if self.matches(TokenType.IDENTIFIER):
            keyword.append(self.definition())

            while self.ignore(TokenType.COMA):
                if self.matches(TokenType.PAR_CLOSE):
                    break
                keyword.append(self.definition())

        self.require(TokenType.PAR_CLOSE)
        return args, vararg, keyword, lazy

    def expression(self):
        data = self.data()
        while self.matches(TokenType.DOT, TokenType.PAR_OPEN, TokenType.BRACKET_OPEN):
            if self.matches(TokenType.DOT):
                self.advance()
                name = self.require(TokenType.IDENTIFIER)
                data = GetAttribute(data, name)
            elif self.matches(TokenType.BRACKET_OPEN):
                self.advance()

                args, coma = [self.expression()], False
                while self.ignore(TokenType.COMA):
                    coma = True
                    if self.matches(TokenType.BRACKET_CLOSE):
                        break
                    args.append(self.expression())

                self.require(TokenType.BRACKET_CLOSE)
                data = GetItem(data, args, coma)
            else:
                data = Call(data, *self.params())

        return data

    def definition(self):
        name = self.require(TokenType.IDENTIFIER)
        self.require(TokenType.EQUALS)
        return Definition(name, self.expression())

    def pair(self):
        key = self.expression()
        self.require(TokenType.COLON)
        return key, self.expression()

    # TODO: unify dict and list?
    def dictionary(self):
        dict_begin = self.require(TokenType.DICT_OPEN)
        pairs = []
        # empty dict
        if self.ignore(TokenType.DICT_CLOSE):
            return Dictionary(pairs, dict_begin)

        pairs.append(self.pair())

        while self.ignore(TokenType.COMA):
            if self.matches(TokenType.DICT_CLOSE):
                break
            pairs.append(self.pair())

        self.require(TokenType.DICT_CLOSE)
        return Dictionary(pairs, dict_begin)

    def array(self):
        array_begin = self.require(TokenType.BRACKET_OPEN)
        values = []
        # empty array
        if self.ignore(TokenType.BRACKET_CLOSE):
            return Array(values, array_begin)

        values.append(self.expression())

        while self.ignore(TokenType.COMA):
            if self.matches(TokenType.BRACKET_CLOSE):
                break
            values.append(self.expression())

        self.require(TokenType.BRACKET_CLOSE)
        return Array(values, array_begin)

    def dotted(self):
        result = [self.require(TokenType.IDENTIFIER)]
        while self.ignore(TokenType.DOT):
            result.append(self.require(TokenType.IDENTIFIER))
        return result

    def import_as(self, allow_dotted):
        if allow_dotted:
            value = self.dotted()
        else:
            value = self.require(TokenType.IDENTIFIER),
        name = None
        if self.ignore(TokenType.AS):
            name = self.require(TokenType.IDENTIFIER)
        return tuple(value), name

    def import_(self):
        root, relative = [], False
        if self.ignore(TokenType.FROM):
            # TODO: I guess this should become legacy
            if self.matches(TokenType.STRING):
                root = self.advance()
                token = self.require(TokenType.IMPORT)
                return ImportPath(root, self.paths(), token)

            relative = self.ignore(TokenType.DOT)
            root = self.dotted()

        main_token = self.require(TokenType.IMPORT)

        # import by path
        # TODO: legacy too
        if self.matches(TokenType.STRING) or self.matches(TokenType.STRING, shift=1):
            return ImportPath(root, self.paths(), main_token)

        if self.ignore(TokenType.ASTERISK):
            return ImportStarred(root, relative)

        # TODO: support
        if relative:
            self.throw("Relative imports are only allowed for config files", main_token)

        block = self.ignore(TokenType.PAR_OPEN)
        values = [self.import_as(not root)]
        while self.ignore(TokenType.COMA):
            values.append(self.import_as(not root))

        if block:
            self.require(TokenType.PAR_CLOSE)
        return ImportPython(root, values, relative, main_token)

    def paths(self):
        block = self.ignore(TokenType.PAR_OPEN)

        paths = [self.require(TokenType.STRING)]
        self.ignore(TokenType.COMA)
        while self.matches(TokenType.STRING):
            paths.append(self.advance())
            self.ignore(TokenType.COMA)

        if block:
            self.require(TokenType.PAR_CLOSE)
        return paths

    def parse(self):
        parents, imports = [], []
        while self.matches(TokenType.IMPORT, TokenType.FROM):
            import_ = self.import_()
            if type(import_) is ImportPython:
                imports.append(import_)
            else:
                if imports:
                    self.throw('Starred and path imports are only allowed at the top of the config', import_.main_token)
                parents.append(import_)

        definitions = []
        while self.position < len(self.tokens):
            definitions.append(self.definition())

        return definitions, parents, imports

    def advance(self):
        result = self.current
        self.position += 1
        return result

    @property
    def current(self):
        if self.position >= len(self.tokens):
            raise SyntaxError('Unexpected end of source')
        return self.tokens[self.position]

    def matches(self, *types, shift=0):
        try:
            temp = self.tokens[self.position + shift]
        except IndexError:
            return False

        for tokenType in types:
            if temp.type == tokenType:
                return True

        return False

    def throw(self, message, token):
        raise SyntaxError(message + '\n  at {}:{}'.format(token.line, token.column))

    def require(self, *types):
        if not self.matches(*types):
            self.throw('Unexpected token: "%s"' % self.current.body, self.current)
        return self.advance()

    def ignore(self, *types):
        if self.matches(*types):
            self.advance()
            return True
        return False


def parse(source, source_path):
    try:
        tokens = tokenize(source)
        for token in tokens:
            token.set_source(source_path)
        return Parser(tokens).parse()
    except SyntaxError as e:
        raise SyntaxError('{} in {}'.format(e.msg, source_path)) from None


def parse_file(config_path):
    with open(config_path) as file:
        return parse(file.read(), config_path)


def parse_string(source):
    return parse(source, '<string input>')



from typing import List

from .structures import Structure
from .token import Token


class Lambda(Structure):
    def __init__(self, params: List[Token], expression: Structure, main_token):
        super().__init__(main_token)
        self.expression = expression
        self.params = params

    def to_str(self, level):
        return 'lambda ' + ','.join(x.body for x in self.params) + ': ' + self.expression.to_str(level + 1)


class Resource(Structure):
    def __init__(self, name: Token):
        super().__init__(name)
        self.name = name

    def to_str(self, level):
        return self.name.body


class GetAttribute(Structure):
    def __init__(self, target: Structure, name: Token):
        super().__init__(name)
        self.target = target
        self.name = name

    def to_str(self, level):
        return '{}.{}'.format(self.target.to_str(level), self.name.body)

    def error_message(self):
        return 'getting attribute {} from {}'.format(self.name.body, self.target.to_str(0))


class GetItem(Structure):
    def __init__(self, target: Structure, args: list, trailing_coma: bool):
        super().__init__(target.main_token)
        self.target = target
        self.args = args
        self.trailing_coma = trailing_coma

    def to_str(self, level):
        result = self.target.to_str(level) + '['
        if not self.args:
            return result + ']'
        else:
            result += '\n'

        for arg in self.args:
            result += '    ' * (level + 1) + arg.to_str(level + 1) + ',\n'
        if not self.trailing_coma:
            result = result[:-2] + '\n'

        return result + '    ' * level + ']'

    def error_message(self):
        return 'getting item from the resource %s' % self.target.to_str(0)


class Call(Structure):
    def __init__(self, target: Structure, args: list, vararg: list, params: list, lazy: bool):
        super().__init__(target.main_token)
        self.target = target
        self.args = args
        self.varargs = vararg
        self.params = params
        self.lazy = lazy

    def to_str(self, level):
        target = self.target.to_str(level)
        body = []
        lazy = ''
        if self.lazy:
            lazy = '    ' * (level + 1) + '# lazy\n'

        for vararg, arg in zip(self.varargs, self.args):
            prefix = '    ' * (level + 1)
            if vararg:
                prefix += '*'
            body.append(prefix + arg.to_str(level + 1))

        for param in self.params:
            body.append('    ' * (level + 1) + param.to_str(level + 1))

        body = lazy + ',\n'.join(body)
        if body:
            body = '\n' + body + '\n' + '    ' * level

        return target + '(' + body + ')'

    def error_message(self):
        return 'calling the resource %s' % self.target.to_str(0)


class Module(Structure):
    def __init__(self, module_type, module_name):
        super().__init__(module_type)
        self.module_type = module_type
        self.module_name = module_name

    def to_str(self, level):
        return '{}:{}'.format(self.module_type.body, self.module_name.body)

    def error_message(self):
        return 'looking for the module %s' % self.to_str(0)


class Literal(Structure):
    def __init__(self, value):
        super().__init__(value)
        self.value = value

    def to_str(self, level):
        return self.value.body


class Array(Structure):
    def __init__(self, values: list, main_token):
        super().__init__(main_token)
        self.values = values

    def to_str(self, level):
        result = '[\n'
        for value in self.values:
            result += '    ' * (level + 1) + value.to_str(level + 1) + ',\n'
        return result + '    ' * level + ']'


class Dictionary(Structure):
    def __init__(self, pairs: list, main_token):
        super().__init__(main_token)
        self.pairs = pairs

    def to_str(self, level):
        result = '{\n'
        for key, value in self.pairs:
            result += '    ' * (level + 1) + '{}: {},\n'.format(key.to_str(level + 1), value.to_str(level + 1))
        return result[:-1] + '    ' * level + '\n}'



from .token import *


def tokenize(source: str):
    tokens = []
    for line_number, line in enumerate(source.splitlines(), 1):
        line = line.rstrip()
        text = line.lstrip()

        while text:
            position = len(line) - len(text) + 1

            # comment or lazy
            if text.startswith('#'):
                text = text.strip()
                match = LAZY.match(text)
                if match:
                    token = Token(match.group(), TokenType.LAZY)
                else:
                    break
            else:
                token = next_token(text)

                if token is None:
                    err = text.split()[0]
                    raise SyntaxError('Unrecognized token: "{}" at {}:{}'.format(err, line_number, position))

            token.add_info(line_number, position)
            tokens.append(token)

            text = text[len(token.body):]
            text = text.strip()
    return tokens


def next_token(text: str):
    for tokenType, regex in REGEXPS.items():
        match = regex.match(text)
        if match:
            match = match.group()
            if match in RESERVED:
                return Token(match, RESERVED[match])
            if match in LITERALS:
                return Token(match, TokenType.LITERAL)
            return Token(match, tokenType)

    for char, tokeType in SINGLE.items():
        if text.startswith(char):
            return Token(char, tokeType)



from collections import defaultdict

from resource_manager.helpers import Scope
from .structures import *


class SyntaxTree:
    def __init__(self, resources: dict):
        self.resources = resources
        self._request_stack = []
        self.cycles = defaultdict(set)
        self.undefined = defaultdict(set)
        self._duplicate_arguments = defaultdict(list)

        self._scopes = []
        self._global = {}
        self._structure_types = []
        for name, node in resources.items():
            self._analyze_tree(name)

    @staticmethod
    def analyze(scope: Scope):
        tree = SyntaxTree(scope._undefined_resources)
        message = ''
        if tree.cycles:
            message += 'Cyclic dependencies found in the following resources:\n'
            for source, cycles in tree.cycles.items():
                message += '  in %s\n    ' % source
                message += '\n    '.join(cycles)
                message += '\n'
        if tree.undefined:
            if message:
                message += '\n'
            message += 'Undefined resources found:\n'
            for source, undefined in tree.undefined.items():
                message += '  in %s\n    ' % source
                message += ', '.join(undefined)
                message += '\n'
        if tree._duplicate_arguments:
            if message:
                message += '\n'
            message += 'Duplicate arguments in lambda definition:\n'
            for source, nodes in tree._duplicate_arguments.items():
                message += '  in %s\n    at ' % source
                message += ', '.join('%d:%d' % node.position()[:2] for node in nodes)
                message += '\n'
        if message:
            raise RuntimeError(message)

    def _analyze_tree(self, name):
        self._request_stack.append(name)
        self._global[name] = False
        self._analyze_node(self.resources[name])
        self._global[name] = True
        self._request_stack.pop()

    def _analyze_node(self, node: Structure):
        node.render(self)

    def _render_resource(self, node: Resource):
        name = node.name.body
        # is it an argument?
        for scope in reversed(self._scopes):
            if name in scope:
                return

        source = node.source()
        # undefined variable:
        if name not in self._global:
            self.undefined[source].add(name)
            return
        # cycle
        if not self._global[name]:
            prefix = " -> ".join(self._request_stack)
            self.cycles[source].add('{} -> {}'.format(prefix, name))

    def _render_get_attribute(self, node: GetAttribute):
        self._analyze_node(node.target)

    def _render_get_item(self, node: GetItem):
        self._analyze_node(node.target)
        for arg in node.args:
            self._analyze_node(arg)

    def _render_call(self, node: Call):
        for arg in node.args:
            self._analyze_node(arg)
        for param in node.params:
            self._analyze_node(param.value)

    def _render_array(self, node: Array):
        for x in node.values:
            self._analyze_node(x)

    def _render_dictionary(self, node: Dictionary):
        for key, value in node.pairs:
            self._analyze_node(key)
            self._analyze_node(value)

    def _render_lambda(self, node: Lambda):
        names = {x.body for x in node.params}
        if len(names) != len(node.params):
            self._duplicate_arguments[node.source()].append(node)
        self._scopes.append(names)
        self._analyze_node(node.expression)
        self._scopes.pop()

    def _render_lazy_import(self, node: LazyImport):
        pass

    def _render_literal(self, node: Literal):
        pass

    def _render_module(self, node: Module):
        pass



import os
from typing import List

from .structures import Structure, Token


class Definition(Structure):
    def __init__(self, name: Token, value: Structure):
        super().__init__(name)
        self.name = name
        self.value = value

    def to_str(self, level):
        return '{} = {}'.format(self.name.body, self.value.to_str(level))


class ImportPython(Structure):
    def __init__(self, root: List[Token], values: list, relative: bool, main_token):
        super().__init__(main_token)
        self.from_ = root
        self.root = '.'.join(x.body for x in root)
        self._values = values
        self.values = [('.'.join(x.body for x in value), name) for value, name in values]
        self.relative = relative

    def to_str(self, level):
        result = ''
        if self.root:
            result += 'from %s ' % self.root
        result += 'import '
        for value, name in self.values:
            result += value + ' '
            if name is not None:
                result += 'as ' + name.body
            result += ', '
        return result[:-2] + '\n'


class ImportStarred(Structure):
    def __init__(self, root: List[Token], relative: bool):
        super().__init__(root[0])

        self.root = root
        self.relative = relative

    def get_paths(self):
        parent = ''
        root = self.root
        if not self.relative:
            parent = root[0].body + ':'
            root = root[1:]
        return [parent + os.sep.join(x.body for x in root) + '.config']

    def to_str(self, level):
        result = 'from '
        if self.relative:
            result += '.'
        return result + '.'.join(x.body for x in self.root) + ' import *'


class ImportPath(Structure):
    def __init__(self, root, paths, main_token):
        super().__init__(main_token)
        self.root = root
        self.paths = paths

    def get_paths(self):
        paths = [eval(x.body) for x in self.paths]
        if not self.root:
            return paths
        return [os.path.join(self.root.body, x) for x in paths]

    def to_str(self, level):
        result = ''
        if self.root:
            result = 'from ' + self.root.body
        return result + 'import ' + ','.join(x.body for x in self.paths)


class LazyImport(Structure):
    def __init__(self, from_, what, as_, relative: bool, main_token: Token):
        super().__init__(main_token)
        self.from_, self.what, self.as_ = from_, what, as_
        self.relative = relative

    def to_str(self, level):
        result = ''
        if self.from_:
            result += 'from '
            if self.relative:
                result += '.'
            result += self.from_ + ' '

        result += 'import %s' % self.what
        if self.as_:
            result += ' as %s' % self.as_.body
        return result + '\n'

    def error_message(self):
        result = 'importing '
        if self.from_:
            result += self.from_ + '.'
        return result + self.what



import re

from .token import Token

first_cap = re.compile('(.)([A-Z][a-z]+)')
all_cap = re.compile('([a-z0-9])([A-Z])')


def snake_case(name):
    name = first_cap.sub(r'\1_\2', name)
    return all_cap.sub(r'\1_\2', name).lower()


class Structure:
    def __init__(self, main_token: Token):
        self.main_token = main_token

    def render(self, walker):
        name = snake_case(self.__class__.__name__)
        return getattr(walker, '_render_' + name)(self)

    def to_str(self, level):
        raise NotImplementedError

    def error_message(self):
        return 'building the resource ' + self.to_str(0)

    def position(self):
        return self.main_token.line, self.main_token.column, self.main_token.source

    def source(self):
        return self.main_token.source


from .expressions import *
from .statements import *



import re
from enum import Enum, auto
from tokenize import Number, String


class TokenType(Enum):
    STRING = auto()
    NUMBER = auto()
    LITERAL = auto()
    IDENTIFIER = auto()
    LAMBDA = auto()

    COLON = auto()
    COMA = auto()
    EQUALS = auto()
    DOT = auto()
    BRACKET_OPEN = auto()
    BRACKET_CLOSE = auto()
    DICT_OPEN = auto()
    DICT_CLOSE = auto()

    LAZY = auto()
    IMPORT = auto()
    FROM = auto()
    AS = auto()
    ASTERISK = auto()

    PAR_OPEN = auto()
    PAR_CLOSE = auto()


REGEXPS = {
    TokenType.NUMBER: re.compile(Number),
    TokenType.STRING: re.compile(String, flags=re.UNICODE),
    TokenType.IDENTIFIER: re.compile(r'[^\d\W]\w*'),
}

RESERVED = {
    'import': TokenType.IMPORT,
    'as': TokenType.AS,
    'from': TokenType.FROM,
    'lambda': TokenType.LAMBDA,
}

LAZY = re.compile(r'^#\s*lazy\s*$')
LITERALS = ('None', 'True', 'False')
SINGLE = {
    ',': TokenType.COMA,
    '*': TokenType.ASTERISK,
    ':': TokenType.COLON,
    '.': TokenType.DOT,
    '=': TokenType.EQUALS,
    '[': TokenType.BRACKET_OPEN,
    ']': TokenType.BRACKET_CLOSE,
    '{': TokenType.DICT_OPEN,
    '}': TokenType.DICT_CLOSE,
    '(': TokenType.PAR_OPEN,
    ')': TokenType.PAR_CLOSE,
}


class Token:
    def __init__(self, body, type, line=None):
        self.body = body
        self.type = type
        self.line = line
        self.column = None
        self.source = None

    def add_info(self, line, column):
        self.line, self.column = line, column

    def set_source(self, source):
        self.source = source



import hashlib
import json
import os
import re

first_cap = re.compile('(.)([A-Z][a-z]+)')
all_cap = re.compile('([a-z0-9])([A-Z])')


def snake_case(name):
    name = first_cap.sub(r'\1_\2', name)
    return all_cap.sub(r'\1_\2', name).lower()


def walk(path, source, exclude):
    modules = []
    if path in exclude:
        return modules

    for root, dirs, files in os.walk(path):
        for directory in dirs:
            dir_path = os.path.join(root, directory)
            modules.extend(walk(dir_path, '{}.{}'.format(source, directory), exclude))

        for file in files:
            name, extension = os.path.splitext(file)
            if extension == '.py':
                file_path = os.path.join(root, file)
                modules.append((file_path, '{}.{}'.format(source, name)))
        break
    return modules


def handle_corruption(db_path):
    db_path = os.path.realpath(db_path)
    raise RuntimeError('Resources base file corrupted. You may want to delete it: ' + db_path) from None


def read_config(path):
    try:
        with open(path, 'r') as file:
            config = json.load(file)
        try:
            hashes = config['hashes']
            config = config['config']
        except KeyError:
            handle_corruption(path)

    except FileNotFoundError:
        config = []
        hashes = {}

    return config, hashes


def get_hash(path, buffer_size=65536):
    current_hash = hashlib.md5()

    with open(path, 'rb') as file:
        data = file.read(buffer_size)
        while data:
            current_hash.update(data)
            data = file.read(buffer_size)

    return '{}_{}'.format(current_hash.hexdigest(), os.path.getsize(path))



from collections import OrderedDict

from .structures import Structure


class Scope:
    def __init__(self):
        self._defined_resources = OrderedDict()
        self._undefined_resources = OrderedDict()
        self._upper = None

    def set_upper(self, upper):
        assert self._upper is None
        self._upper = upper

    def overwrite(self, scope):
        assert not self._defined_resources and not scope._defined_resources
        for name, value in scope._undefined_resources.items():
            self._undefined_resources[name] = value

    def set_resource(self, name: str, value: Structure):
        if name in self._undefined_resources:
            raise SyntaxError('Duplicate definition of resource "%s" in %s' % (name, value.source()))
        self._undefined_resources[name] = value

    def define_resource(self, name: str, value):
        if name in self._defined_resources:
            raise SyntaxError('Duplicate definition of resource "%s" in %s' % (name, value.source()))
        self._defined_resources[name] = value

    def get_resource(self, name: str, renderer):
        if name in self._defined_resources:
            return self._defined_resources[name]

        if name not in self._undefined_resources:
            if self._upper is None:
                raise AttributeError('Resource "{}" is not defined'.format(name))
            return self._upper.get_resource(name, renderer)

        node = self._undefined_resources[name]
        resource = renderer(node)
        self._defined_resources[name] = resource
        return resource


class LambdaFunction:
    def __init__(self, node, scope, interpreter):
        self.node = node
        self.scope = scope
        self.interpreter = interpreter

    def __call__(self, *args):
        scope = Scope()
        for x, y in zip(self.node.params, args):
            scope.define_resource(x.body, y)
        scope.set_upper(self.scope)

        self.interpreter._scopes.append(scope)
        try:
            result = self.interpreter._render(self.node.expression)
        finally:
            self.interpreter._scopes.pop()
        return result



from distutils.core import setup

classifiers = '''Development Status :: 5 - Production/Stable
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6'''

with open('README.rst', encoding='utf-8') as file:
    long_description = file.read()

setup(
    name='resource-manager',
    packages=['resource_manager'],
    include_package_data=True,
    version='0.6.1',
    description='A small resource manager for config files',
    long_description=long_description,
    author='maxme1',
    author_email='maxs987@gmail.com',
    license='MIT',
    url='https://github.com/maxme1/resource-manager',
    download_url='https://github.com/maxme1/resource-manager/archive/0.6.1.tar.gz',
    keywords=[
        'config', 'manager', 'resource'
    ],
    classifiers=classifiers.splitlines(),
    install_requires=[]
)
